from .._regexes import match_quality_int

__all__ = ('Quality',)


class Quality(str):
    __slots__ = ('_i', 'addon')
    addon: str  # can contain 'ultra'

    def __init__(self, *_, **__):
        _i = match_quality_int(self)
        if not _i:
            raise ValueError(f'{self!r} is not quality.')
        _i, self.addon = _i.groups()
        self.addon = self.addon.casefold()
        self._i = int(_i)

    def __int__(self):
        """
        returns pixels height
        """
        return self._i

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return super().__le__(other)
        if not (self.addon or other.addon):
            return self._i < int(other)
        return False



from typing import Iterable, TypeVar

from ._dataclass import frozen_slots_dataclass
from .._bs4 import BeautifulSoup
from ..api.http import get_response
from ..errors import EmptyPage
from ..stream.player import *

__all__ = ('Page', 'PageNumber', 'InlineItem')

T = TypeVar('T')
PageNumber = TypeVar('PageNumber', int, slice, Iterable[int])


def _range_from_slice(obj: slice | T) -> range | T:
    if isinstance(obj, slice):
        return range(*[i for i in (obj.start, obj.stop, obj.step) if i is not None])
    return obj


@frozen_slots_dataclass
class InlineItem:
    """Content Inline Item view"""
    url: str
    name: str
    info: str
    poster: str
    'Image url'

    @property
    async def player(self) -> PlayerType:
        """Return a Player Instance"""
        return await Player(self.url)


class Page:
    """AJAX class for HDRezka search"""
    __slots__ = ('_page', '_page_format', '__yields', '__yields_page')

    def __init__(self, url: str = 'https://rezka.ag/'):
        self.__yields = []
        self.__yields_page = 0
        self.page = url

    @property
    def page(self) -> str:
        return self._page

    @page.setter
    def page(self, value):
        if not isinstance(value, str):
            value = str(value)
        # noinspection HttpUrlsUsage
        if not (value.startswith('https://') or value.startswith('http://')):
            value = f'https://{value}'
        self._page = value
        self._page_format = self._concat_paginator(value.removesuffix('/')).format

    @staticmethod
    def _concat_paginator(url: str) -> str:
        return f'{url}/page/{{0}}'

    def __aiter__(self):
        self.__yields_page = 0
        self.__yields.clear()
        return self

    async def __anext__(self) -> InlineItem:
        if not self.__yields:
            try:
                self.__yields_page += 1
                self.__yields += await self.get_page(self.__yields_page)
            except EmptyPage:
                raise StopAsyncIteration
        return self.__yields.pop(0)

    async def get_page(self, page: PageNumber) -> list[InlineItem]:
        """
        Get items from pages range.
        None for all elements of all pages
        """
        items = BeautifulSoup((await get_response('GET', self._page_format(page))).text
                              ).find_all(class_='b-content__inline_item')
        if not items:
            raise EmptyPage(page)
        return [InlineItem(
            (a := (link := i.find(class_='b-content__inline_item-link')).find('a'))['href'],
            a.text,
            link.find('div').text,
            i.find(class_='b-content__inline_item-cover').find('img').get('src', ''))
            for i in items]

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.page!r})'



from collections import defaultdict
from typing import Any, SupportsInt, TypeVar

from .._bs4 import BeautifulSoup
from ..api.ajax import AJAX
from ..errors import UnknownContentType
from ..post import *

__all__ = ('Player', 'PlayerType', 'PlayerBase', 'PlayerMovie', 'PlayerSeries')


class _CacheStorage:
    __slots__ = ('max_size', '__storage')

    def __init__(self, max_size: int = 10):
        self.max_size = max_size
        self.__storage = {}

    def get(self, key):
        return self.__storage.get(key)

    def set(self, key, value):
        self.__storage[key] = value
        for k in (*self.__storage.keys(),):
            if len(self.__storage) <= self.max_size:
                break
            del self.__storage[k]


__CACHE = _CacheStorage()


class PlayerBase:
    __slots__ = ('post',)

    def __init__(self, url_or_cast: Any):
        """Need await"""
        if isinstance(url_or_cast, PlayerBase):
            self.post = url_or_cast.post
            return
        elif not isinstance(url_or_cast, str):
            url_or_cast = str(url_or_cast)
        self.post = Post(url_or_cast)

    def __await__(self):
        yield from self.post.__await__()

    def _translator(self, translator_id: SupportsInt = None) -> int:
        if translator_id is None:
            return self.post.translator_id
        translator_id = int(translator_id)
        return self.post.translators.ids[abs(translator_id)] if translator_id <= 0 else translator_id

    def __repr__(self):
        return f"{self.__class__.__qualname__}({short_url(self.post.url)!r})"


class PlayerMovie(PlayerBase):
    __slots__ = ()

    async def get_stream(self, translator_id: SupportsInt = None) -> URLs:
        return urls_from_ajax_response(await AJAX.get_movie(self.post.id, self._translator(translator_id)))


class PlayerSeries(PlayerBase):
    __slots__ = ()

    async def get_episodes(self, translator_id: SupportsInt = None) -> defaultdict[int, tuple[int]]:
        episodes = BeautifulSoup((await AJAX.get_episodes(self.post.id, self._translator(translator_id)))['episodes'])
        result: defaultdict[int, tuple[int]] = defaultdict(tuple)
        for i in episodes.find_all(class_='b-simple_episode__item', attrs=('data-season_id', 'data-episode_id')):
            result[int(i.get('data-season_id'))] += int(i.get('data-episode_id')),
        return result

    async def get_stream(self, season: int, episode: int, translator_id: SupportsInt = None) -> URLs:
        return urls_from_ajax_response(
            await AJAX.get_stream(self.post.id, self._translator(translator_id), season, episode))


PlayerType = TypeVar('PlayerType', PlayerBase, PlayerMovie, PlayerSeries)


async def player(url_or_path: Any) -> PlayerType:
    """
    Returns either Player Series if series, or PlayerMovie if movie, otherwise raises UnknownContentType
    """
    cast = PlayerBase(url_or_path)
    cached = __CACHE.get(cast.post.url)
    if cached is not None:
        return cached
    await cast
    type = cast.post.type
    match type:
        case 'tv_series':
            value = PlayerSeries(cast)
        case 'movie':
            value = PlayerMovie(cast)
        case _:
            raise UnknownContentType(type)
    __CACHE.set(cast.post.url, value)
    return value


Player = player



from .._regexes import findall_subtitles
from ..._dataclass import frozen_slots_dataclass

__all__ = ('SubtitleURL', 'SubtitleURLs')


@frozen_slots_dataclass
class SubtitleURL:
    url: str
    # language:
    name: str
    code: str


class SubtitleURLs:
    __slots__ = ('subtitles', 'has_subtitles', 'subtitle_names', 'subtitle_codes', 'default')

    def __init__(self, subtitle: str | bool, subtitle_lns: dict[str, str] | bool, subtitle_def: str | bool):
        self.has_subtitles = not subtitle
        if not subtitle_lns:
            subtitle_lns = {'off': ''}
        self.subtitles: tuple[SubtitleURL] = *(
            SubtitleURL(url, name, subtitle_lns[name]) for name, url in findall_subtitles(subtitle or '')),
        self.subtitles += SubtitleURL('', '', 'off'),
        self.subtitle_names = {}
        self.subtitle_codes = {}
        for subtitle in self.subtitles:
            self.subtitle_names[subtitle.name] = self.subtitle_codes[subtitle.code] = subtitle

        self.default: SubtitleURL | None = self.subtitle_codes.get(subtitle_def)

    def __getitem__(self, item: str) -> SubtitleURL | None:
        return self.subtitle_names.get(item, self.subtitle_codes.get(item))

    def __getattr__(self, item: str) -> SubtitleURL:
        return self.subtitle_codes[item]

    def __bool__(self):
        return self.has_subtitles

    def __repr__(self):
        return f'{self.__class__.__qualname__}<{self.subtitles!r}>'



import re

__all__ = ('findall_qualities', 'findall_subtitles', 'shorten_url_match', 'match_quality_int')

findall_qualities = re.compile(r'\[([^]]+)](\S+)(?:\sor\s|$)').findall
findall_subtitles = re.compile(r'\[([^]]+)]([^,]+)(?:,|$)').findall
shorten_url_match = re.compile(r'(?:(?:https?://)?rezka\.ag/)?\D*(\d+)\S*(?:\.html)?/?', re.I).match
match_quality_int = re.compile(r'(\d+)[pi]\s*($|\w+)').match



from typing import Iterable, SupportsInt

from .quality import Quality
from .._regexes import findall_qualities
from ...._antiobfuscation import clear_trash

__all__ = ('VideoURL', 'VideoURLs')


class VideoURL(str):
    __slots__ = ('mp4',)

    def __init__(self, *_, **__):
        self.mp4 = str(self.removesuffix('8').removesuffix(':hls:manifest.m3u'))


class VideoURLs:
    __slots__ = ('raw_data', 'qualities', 'min')

    def __init__(self, data: str | dict):
        if isinstance(data, str):
            self.raw_data: dict[Quality, VideoURL] = {
                Quality(q): VideoURL(u) for q, u in findall_qualities(clear_trash(data))}
        elif isinstance(data, dict):
            self.raw_data = data
        else:
            raise TypeError(f'got {data!r} (type {type(data)}) but expected type {str | dict}')
        self.qualities: tuple[Quality] = *sorted(self.raw_data),
        self.min = int(self.qualities[0]) if self.qualities else 1

    @property
    def last_url(self) -> VideoURL:
        return self[-1].raw_data.popitem()[1]

    def __getitem__(self, item: str | SupportsInt | Iterable | slice):
        isiter = False
        if not isinstance(item, str):
            if isinstance(item, slice):
                item = item.start, item.stop, item.step
            isiter = isinstance(item, Iterable) and item
        if isiter:
            result = self
            for part in item:
                if part is not None:
                    result = result[part]
            result = result.raw_data
        elif isinstance(item, int):
            if item < self.min:
                item = self.qualities[item]
                result = {item: self.raw_data[item]}
            else:
                result = {q: v for q, v in self.raw_data.items() if int(q) == item}
        elif isinstance(item, str):
            item = item.casefold()
            result = {q: v for q, v in self.raw_data.items() if q.addon == item}
        else:
            raise TypeError(f'Invalid type {type(item)}')
        return self.__class__(result)

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.raw_data})"



from .._dataclass import frozen_slots_dataclass

__all__ = ('Rating', 'Place', 'Release', 'AgeRating', 'Duration', 'Poster')


@frozen_slots_dataclass
class Rating:
    service: str
    rating: int | float
    votes: int = 0


@frozen_slots_dataclass
class Place:
    name: str
    place: int


@frozen_slots_dataclass
class Release:
    year: int
    day: str


@frozen_slots_dataclass
class AgeRating:
    age: int
    description: str


@frozen_slots_dataclass
class Duration:
    number: int
    units: str


@frozen_slots_dataclass
class Poster:
    full: str
    preview: str



from functools import lru_cache

from ._regexes import shorten_url_match

__all__ = ('short_url', 'long_url')


@lru_cache(1024)
def short_url(url: str) -> str:
    """
    Returns string rezka.ag post with format "{id}-{id}" (valid path)
    """
    parts = shorten_url_match(url)
    if parts is None:
        return url
    id = parts.group(1)
    return f'{id}-{id}'


@lru_cache(1024)
def long_url(url: str) -> str:
    """
    Returns full url of rezka.ag post
    """
    return f'https://rezka.ag/{short_url(url)}.html'



from .kind import SubtitleURLs, VideoURLs
from .._dataclass import frozen_slots_dataclass


@frozen_slots_dataclass
class URLs:
    video: VideoURLs
    subtitles: SubtitleURLs


def urls_from_ajax_response(response: dict[str]) -> URLs:
    return URLs(VideoURLs(response.get('url', '')),
                SubtitleURLs(response.get('subtitle', False),
                             response.get('subtitle_lns', False), response.get('subtitle_def', False)))



r"""
# [HDRezka](https://rezka.ag/) site API.

# Install

`pip install HDRezka`

# Example

```python
import asyncio
from hdrezka import Search

async def main():
    player = await (await Search('Breaking Bad').get_page(1))[0].player  # or just use await Player(646)
    print(player.post.info, end='\n\n')

    translator_id = None  # default
    for name, id_ in player.post.translators.name_id.items():
        if 'субтитры' in name.casefold(): translator_id = id_; break

    stream = await player.get_stream(1, 1, translator_id)  # raise AJAXFail if invalid episode or translator
    video = stream.video
    print(video.last_url)  # best quality (.m3u8)
    print(video[video.min].last_url.mp4, end='\n\n')  # worst quality (.mp4)

    subtitles = stream.subtitles
    print(subtitles.default.url)  # subtitles.ru.url or subtitles['Русский'].url

if __name__ == '__main__': asyncio.run(main())
```

# [GitHub](https://github.com/NIKDISSV-Forever/HDRezka)
"""
from .api import *
from .post import *
from .stream import *



import httpx

__all__ = ('get_response', 'DEFAULT_CLIENT_KWARGS', 'DEFAULT_REQUEST_KWARGS')

DEFAULT_CLIENT_KWARGS = {'headers': {'user-agent': 'Mozilla/5.0'}}
DEFAULT_REQUEST_KWARGS = {}


async def get_response(*args, **kwargs):
    async with httpx.AsyncClient(**DEFAULT_CLIENT_KWARGS) as client:
        return await client.request(*args, **kwargs, **DEFAULT_REQUEST_KWARGS)



import re
from inspect import getdoc
from typing import Any

from bs4 import Tag

from .fields import *
from ..urls import Quality
from ..._bs4 import BeautifulSoup

__all__ = ('PostInfo',)
_find_any_digit = re.compile(r'\d+').findall


def _get_int(value: Any, /, default: int = 0) -> int:
    if not isinstance(value, str):
        value = str(value)
    if digits := _find_any_digit(value):
        return int(''.join(digits))
    return default


class PostInfo:
    __slots__ = ('rating', 'places', 'slogan', 'release', 'country', 'director', 'genre',
                 'quality', 'translator', 'age_rating', 'duration', 'from_', 'characters',
                 '_view', 'fields', 'title', 'orig_title', 'poster', 'description')
    FILL_FIELDS: tuple[str] = __slots__[:13]

    def __init__(self, soup: BeautifulSoup):
        post_info = soup.find(class_='b-post__info')
        self.title = soup.find(class_='b-post__title').text.strip()
        orig_title = soup.find(class_='b-post__origtitle')
        self.orig_title = orig_title.text.strip() if orig_title else self.title
        poster = soup.select_one('.b-sidecover>a')
        self.poster = Poster(poster['href'], poster.find('img')['src'])
        self.description = soup.find(class_='b-post__description_text').text.strip()
        self._view = self._form_view(post_info.find_all('tr'))

        raw_fields = post_info.select('tr>td')
        if len(raw_fields) % 2:
            raw_fields.append(raw_fields[-1])

        fields = {raw_fields[i].find('h2').text.strip().removesuffix(':'): raw_fields[-~i]
                  for i in range(0, len(raw_fields), 2)}
        for field in self.FILL_FIELDS:
            meth = getattr(self, f'_get_field_{field}', None)
            field_name = '' if meth is None else (getdoc(meth) or '').strip()
            item = fields.get(field_name)
            if not item:
                setattr(self, field, None)
                continue
            processed = meth(item)
            fields[field_name] = processed
            setattr(self, field.removeprefix('_get_field_'), processed)
        self.fields = fields

    @staticmethod
    def _get_field_rating(tag: Tag) -> dict[str, Rating]:
        """Рейтинги"""
        result = {}
        for span in tag.find_all('span', class_='b-post__info_rates'):
            classes: list[str] = span.get('class')
            classes.remove('b-post__info_rates')
            service = span.find('a').text
            votes = span.find('i')
            if votes:
                votes = _get_int(votes.text)
            result[classes[0] if classes else service] = Rating(
                service=service, rating=float(span.find('span').text), votes=votes)
        return result

    @staticmethod
    def _get_field_places(tag: Tag) -> tuple[Place]:
        """Входит в списки"""
        return *(Place(a.text, _get_int(a.next_element.next_element)) for a in tag.find_all('a')),

    @staticmethod
    def _get_field_slogan(tag: Tag) -> str:
        """Слоган"""
        return tag.get_text(strip=True)

    @staticmethod
    def _get_field_release(tag: Tag) -> Release:
        """Дата выхода"""
        return Release(day=tag.next_element.text.strip(), year=_get_int(tag.find('a').text))

    @staticmethod
    def _get_field_country(tag: Tag) -> tuple[str]:
        """Страна"""
        return *map(str.strip, tag.text.split(',')),

    @staticmethod
    def _get_field_director(tag: Tag) -> str:
        """Режиссер"""
        return tag.get_text(strip=True)

    @staticmethod
    def _get_field_genre(tag: Tag) -> tuple[str]:
        """Жанр"""
        return *map(str.strip, tag.text.split(',')),

    @staticmethod
    def _get_field_quality(tag: Tag):
        """В качестве"""
        return Quality(tag.text.strip())

    @staticmethod
    def _get_field_translator(tag: Tag) -> str:
        """В переводе"""
        return tag.get_text(strip=True)

    @staticmethod
    def _get_field_age_rating(tag: Tag) -> AgeRating:
        """Возраст"""
        return AgeRating(_get_int(span := tag.find('span').next_element), span.next_element.get_text(strip=True))

    @staticmethod
    def _get_field_duration(tag: Tag) -> Duration:
        """Время"""
        return Duration(*tag.text.split(' ', 1))

    @staticmethod
    def _get_field_from_(tag: Tag) -> tuple[str]:
        """Из серии"""
        return *(a.text for a in tag.find_all('a')),

    @staticmethod
    def _get_field_characters(tag: Tag) -> tuple[str]:
        """В ролях актеры"""
        return *(span.text for span in tag.select('span[itemprop="name"]')),

    def _form_view(self, table: list[Tag]) -> str:
        return '{0}\n\n{1}\n\n{2}'.format(
            self.title if self.orig_title == self.title else f'{self.title} ({self.orig_title})',
            '\n'.join(tr.text.replace(')', ') ').strip() for tr in table),
            self.description
        )

    def __str__(self):
        return self._view

    def __repr__(self):
        # noinspection PyUnboundLocalVariable
        return f"""{self.__class__.__qualname__}<{', '.join(
            f'{name}={attr!r}' for name in self.__slots__
            if not name.startswith('_') and (attr := getattr(self, name)))}>"""



from .http import get_response
from ..errors import AJAXFail

__all__ = ('AJAX', 'AnyID')

AnyID = int | str


class AJAX:
    """HDRezka AJAX class"""
    __slots__ = ()

    @classmethod
    async def _send_data(cls, action: str, data: dict[str]) -> dict[str]:
        answer: dict = (await get_response('POST', f'https://rezka.ag/ajax/{action}/', data=data)).json()
        if not answer.get('success', True):
            raise AJAXFail(answer.get('message', answer))
        return answer

    @classmethod
    async def get_cdn_series(cls, data: dict[str]):
        return await cls._send_data('get_cdn_series', data)

    @classmethod
    async def get_episodes(cls, id: AnyID, translator_id: AnyID):
        return await cls.get_cdn_series(
            {'action': 'get_episodes',
             'id': id,
             'translator_id': translator_id})

    @classmethod
    async def get_stream(cls, id: AnyID, translator_id: AnyID, season: AnyID, episode: AnyID) -> dict[str]:
        return await cls.get_cdn_series(
            {'action': 'get_stream',
             'id': id,
             'translator_id': translator_id,
             'season': season,
             'episode': episode}
        )

    @classmethod
    async def get_movie(cls, id: AnyID, translator_id: AnyID):
        return await cls.get_cdn_series(
            {'action': 'get_movie',
             'id': id,
             'translator_id': translator_id}
        )
